import logging
import logging.config
import logging.handlers
import typing as t
import sys


LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'simple': {
            'format': "[Piedmont][%(levelname)s]:\t%(message)s"
        },
        'detailed': {
            'format': "$ %(asctime)s [%(name)s][%(levelname)s][%(threadName)s:%(process)d]::%(module)s::\t%(message)s"
        }
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'formatter': 'simple',
            'stream': sys.stdout
        },
        'file': {
            'class': 'logging.handlers.RotatingFileHandler',
            'level': 'DEBUG',
            'filename': 'piedmont.log',
            'formatter': 'detailed',
            'maxBytes': 10485760,  # 10MB
            'backupCount': 5
        }
    },
    'loggers': {
        'piedmont': {
            'level': 'ERROR',
            'handlers': ['file'],
            'propagate': False
        },
        'piedmont-console': {
            'level': 'ERROR',
            'handlers': ['console'],
            'propagate': False
        }
    }
}


logging.config.dictConfig(LOGGING_CONFIG)

logger = logging.getLogger('piedmont')
console = logging.getLogger('piedmont-console')


def info(msg: object, *args, **kwargs):
    logger.info(msg, *args, **kwargs)
    console.info(msg, *args, **kwargs)


def debug(msg: object, *args, **kwargs):
    logger.debug(msg, *args, **kwargs)
    console.debug(msg, *args, **kwargs)


def warning(msg: object, *args, **kwargs):
    logger.warning(msg, *args, **kwargs)
    console.warning(msg, *args, **kwargs)


def error(msg: object, *args, **kwargs):
    logger.error(msg, *args, **kwargs)
    console.error(msg, *args, **kwargs)


def critical(msg: object, *args, **kwargs):
    logger.critical(msg, *args, **kwargs)
    console.critical(msg, *args, **kwargs)


def set_dev_mode(flag=True):
    if flag:
        logger.setLevel(logging.DEBUG)
        console.setLevel(logging.DEBUG)
        console.debug('Set log level to: `DEBUG`.')
        logger.debug(
            f'\n{"=" * 48}\n{">"*15} PIEDMONT DEV LOG {"<"*15}\n{"=" * 48}'
        )
    else:
        logger.setLevel(logging.CRITICAL)
        console.setLevel(logging.INFO)



class DuplicateHandlerError(Exception):
    def __init__(self, name):
        super().__init__(f"Duplicate handler for message:{name}")


class KeyPathError(Exception):
    def __init__(self, key: str):
        super().__init__(
            f'First key in key path can not be number. Check out the query key: `{key}`')


class StackNotExists(Exception):
    def __init__(self, key: str):
        super().__init__(
            f'The stack with name `{key}` doesn\'t exits.'
        )


class ConfigError(Exception):
    def __init__(self, message):
        super().__init__(message)



import pytest
from piedmont.storage import Storage, _is_valid_number_string


@pytest.fixture(scope='function')
def storage():
    s = Storage()
    s._list['default'] = [f'test{i}' for i in range(5)]
    s._stack['default'] = [f'string{i}' for i in range(5)]
    s._queue['default'] = [f'task{i}' for i in range(5)]
    s._dict = {f'key{i}': f'value{i}' for i in range(5)}
    return s


def test_append_storage_will_add_to_array(storage):
    count = storage.list_count()
    storage.append('test')
    assert storage.list_count() == count + 1


def test_push_object_will_add_to_stack(storage):
    count = storage.stack_count()
    storage.push('test')
    assert storage.stack_count() == count + 1


def test_pop_object_will_remove_last_object_in_stack(storage):
    last = storage.stack()[-1]
    count = storage.stack_count()
    obj = storage.pop()
    assert obj == last
    assert storage.stack_count() == count - 1


def test_peek_will_return_last_object_but_not_remove(storage):
    count = storage.stack_count()
    result = storage.peek()
    assert result == storage.stack()[-1]
    assert storage.stack_count() == count


def test_set_value_by_string_key_chain(storage):
    storage._dict = {
        # 'key1': {
        #     'key2': {}
        # }
    }
    storage.set_value_by_key('key1.key2.key3', 'new value')
    assert storage._dict['key1']['key2']['key3'] == 'new value'


def test_set_value_by_key_chain_with_idx_as_first_key(storage):
    from piedmont.errors import KeyPathError
    with pytest.raises(KeyPathError) as exc_info:
        storage.set_value_by_key('0.key1.key2', 'value')

    assert exc_info.type == KeyPathError


def test_set_value_by_key_chain_with_idx_as_last_key(storage):
    storage._dict = {
        'key1': {
            'key2': [
                1, 2, 3, 4, 5
            ]
        }
    }
    storage.set_value_by_key('key1.key2.3', 100)
    print(storage.dict)
    assert storage.dict['key1']['key2'][3] == 100


def test_set_json_value_by_key_will_convert_to_object(storage):
    storage.set_value_by_key('test.key', '[1,2,3,4,5]')
    assert isinstance(storage.dict['test']['key'], list)
    assert len(storage.dict['test']['key']) == 5


def test_set_value_by_single_key(storage):
    storage.set_value_by_key('test key', 'test_value')
    assert storage.dict['test key'] == 'test_value'


def test_clear_data(storage):
    storage.clear()
    assert storage.dict_count() == 0
    storage.clear('list')
    assert storage.list_count() == 0
    storage.clear('queue')
    assert storage.queue_count() == 0
    storage.clear('stack')
    assert storage.stack_count() == 0


def test_get_value_by_none_digit_index(storage):
    with pytest.raises(ValueError) as exc_info:
        result = storage.get_value_at_index('index')
    assert exc_info.type == ValueError


def test_get_value_by_index(storage):
    result = storage.get_value_at_index(3)
    assert result == storage.list()[3]
    result = storage.get_value_at_index('2')
    assert result == storage.list()[2]


def test_show_data(storage):
    storage.show_dict()
    storage.show_list()
    storage.show_stack()
    storage.show_queue()


def test_get_value_by_single_key(storage):
    result = storage.get_value_by_key('key1')
    assert result == 'value1'


def test_get_value_by_key_chain(storage):
    storage._dict = {
        'key1': {
            'key2': [
                1, 2, 3, 4, 5
            ]
        }
    }

    result = storage.get_value_by_key('key1.key2')
    assert isinstance(result, list)
    assert len(result) == 5
    result = storage.get_value_by_key('key1.key2.3')
    assert result == 4


def test_is_valid_number_string():
    assert _is_valid_number_string('asdf') == False
    assert _is_valid_number_string('123') == True
    assert _is_valid_number_string('-10') == True


def test_append_object_with_key(storage):
    storage.append('test', 'key')
    storage.show_list()
    assert storage.list_count('key') == 1


def test_push_object_with_key(storage):
    storage.push('value1', key='key')
    storage.push('value2', key='key')
    storage.show_stack()
    assert storage.stack_count('key') == 2
    assert storage._stack['key'][1] == 'value2'


def test_remove_from_list(storage):
    storage.clear('list')
    for i in range(5):
        storage.append(f'test{i}')

    storage.remove('test3')
    storage.show_list()
    assert storage.list_count() == 4
    storage.append('new value', 'key')
    assert storage.list_count('key') == 1
    storage.remove('new value', 'key')
    assert storage.list_count('key') == 0
    storage.remove('none exists')
    assert storage.list_count() == 4


def test_append_into_and_pop_from_queue(storage):
    storage.clear('queue')
    for i in range(5):
        storage.push(f'data{i}', 'queue')

    assert storage.queue_count() == 5

    result = storage.pop('queue')
    assert result == 'data0'



from piedmont.config import Config


def test_default_config():
    config = Config()
    assert config is not None
    assert 'APP_NAME' in config.keys()
    assert config.port == '9981'


def test_yaml_config():
    config = Config('tests/config.yaml')
    assert config is not None
    assert 'APP_NAME' in config.keys()
    assert config.port == '9981'


def test_create_with_value():
    config = Config('test')



import pytest
from piedmont.config import Config
from piedmont.logger import logger


def test_load_config():
    config = Config('tests/config.yaml')
    assert config is not None
    assert 'name' in config.bridge_conf.keys()
    logger.debug('Test debug message111')
    logger.info('Test info message.')



from __future__ import annotations

import typing as t
import json
import functools
import socketio
import socketio.exceptions

from .config import Config
from .typing import T_Handler, T_Mapper
from .errors import DuplicateHandlerError, KeyPathError, StackNotExists
from .storage import storage
from . import logger

PP_BRIDGE_APP = 'ppBridgeApp'
PP_MESSAGE = 'ppMessage'

BUILD_IN_MESSAGES = [
    # build in stack commands.
    ['pie.push', '_push'],
    ['pie.pop', '_pop'],
    ['pie.peek', '_peek'],
    # Build in queue commands.
    ['pie.pushq', '_q_push'],
    ['pie.popq', '_q_pop'],
    ['pie.peekq', '_q_peek'],
    # build in list commands.
    ['pie.insert', '_insert'],
    ['pie.append', '_append'],
    ['pie.index', '_index'],
    ['pie.remove', '_remove'],
    # ['pie.update', '_update'],
    # ['pie.sort', '_sort'],
    # ['pie.filter', '_filter'],
    # ['pie.count', '_count'],
    # ['pie.sum', '_sum'],
    # ['pie.avg', '_avg'],
    # build in dict commands.
    ['pie.set', '_set_value_by_key'],
    ['pie.get', '_get_value_by_key'],
    # build in data display commands.
    ['pie.list', '_show_list'],
    ['pie.stack', '_show_stack'],
    ['pie.queue', '_show_queue'],
    ['pie.dict', '_show_dict'],
    # help message.
    ['pie.help', '_show_help'],
    ['pie.clear', '_clear']
]


def _safe_load_key(data: any):
    if isinstance(data, dict) and 'key' in data and 'piedmont.payload' in data:
        return data['key']
    return 'default'


def _safe_load_data(data: any):
    if isinstance(data, dict) and 'value' in data and 'piedmont.payload' in data:
        return data['value']
    return data


def _is_invalid_key(data: any):
    return not isinstance(data, dict) \
        or not 'key' in data \
        or len(data['key']) == 0


def _key_index_parser(data: any):
    key = str(data.get('key', None))
    if key:
        key_path = key.split('.')
        idx = key_path.pop()
        if not idx.isdigit():
            idx = 0
        else:
            idx = int(idx)
            if len(key_path) == 0:
                key = 'default'
            else:
                key = '.'.join(key_path)
    else:
        key = 'default'
        idx = 0
    return key, idx


class Piedmont():

    _client = socketio.Client()
    _config: Config
    _handler_mapper: T_Mapper

    def __init__(
            self,
            config: Config = None,
            debug: bool = False,
            auto_connect: bool = False,
            separator: str = "::"
    ) -> None:
        super().__init__()
        self._handler_mapper = {}
        logger.set_dev_mode(debug)
        self._config = config or Config()
        self.separator = separator
        self._regist_handlers()
        self._regist_buildin_handlers()
        if auto_connect:
            self.connect()

    def start(self):
        try:
            if not self._client.connected:
                self.connect()

            while self._client.connected:
                pass
        except KeyboardInterrupt:
            if self._client.connected:
                self._client.disconnect()

    def _regist_buildin_handlers(self):
        for buildin in BUILD_IN_MESSAGES:
            self._handler_mapper.setdefault(
                buildin[0],
                getattr(self, buildin[1])
            )

    def _set_value_by_key(self, data):
        if _is_invalid_key(data):
            self._error_message(
                f'You need to provide a key for `pie.set`, like `pie.set{self.separator}key`'
            )
            return

        key = data['key']
        value = data['value']
        try:
            result = storage.set_value_by_key(key, value)
            self.send(f'key::{key}', result)
        except KeyPathError as e:
            self._error_message(f'{e}')

    def _get_value_by_key(self, data):
        try:
            result = storage.get_value_by_key(data['key'])
            logger.debug(data)
            self.send(f'key::{data['key']}', result)
        except KeyError as e:
            self._error_message(f'{e}')

    def _show_help(self, data):
        pass

    def _show_stack(self, data):
        self.send('stack.all', storage._stack)

    def _show_dict(self, data):
        self.send('dict.all', storage.dict)

    def _show_list(self, data):
        self.send('list.all', storage._list)

    def _show_queue(self, data):
        self.send('queue.all', storage._queue)

    def error(self, msg: str):
        self.send('pie.error', msg)

    def _push(
            self, data,
            target: t.Literal['stack', 'queue'] = 'stack',
    ):
        logger.debug(f'push data: `{data}`, type of data: {type(data)}')
        result = storage.push(
            _safe_load_data(data),
            target,
            _safe_load_key(data)
        )
        self.send(f'push.{_safe_load_key(data)}', result)

    def _pop(
        self, data,
        target: t.Literal['stack', 'queue'] = 'stack',
    ):
        try:
            key = _safe_load_key(data)
            try:
                idx = int(_safe_load_data(data))
            except ValueError:
                idx = None
            except TypeError:
                idx = None
            result = storage.pop(target, key, idx)
            self.send(f'pop.{key}', result)
        except IndexError:
            logger.info(f'{target}.{key} storage is empty.')
            self.send(f'pop.{key}.empty')
        except StackNotExists as e:
            self._error_message(f'{e}')

    def _peek(
        self, data,
        target: t.Literal['stack', 'queue'] = 'stack',
        key: str = 'default'
    ):
        try:
            result = storage.peek(target, _safe_load_key(data))
            self.send(f'peek.{_safe_load_key(data)}', result)
        except IndexError:
            logger.info(f'{target}.{_safe_load_key(data)} storage is empty.')
            self.send(f'peek.{_safe_load_key(data)}.empty')

    def _q_push(self, data):
        self._push(data, 'queue')

    def _q_pop(self, data):
        self._pop(data, 'queue')

    def _q_peek(self, data):
        self._peek(data, 'queue')

    def _append(self, data):
        logger.debug(f'- ACTION: append data: `{data}`')
        result = storage.append(
            _safe_load_data(data),
            _safe_load_key(data)
        )
        self.send(f'append.{_safe_load_key(data)}', result)
        logger.debug(f'{storage._list}')

    def _insert(self, data):
        logger.debug(f'- ACTION: insert data: `{data}`')
        if isinstance(data, dict):
            value = data.get('value', None)
            key, idx = _key_index_parser(data)
        else:
            value = data
            idx = 0
            key = 'default'

        logger.debug(f'value type: {type(value)}')

        if not value:
            self._error_message(
                f'You must provide a valid value for `pie.insert`')
            return

        result = storage.insert(value, idx, key)
        self.send(f'insert.{key}', result)
        logger.debug(f'{storage._list}')

    def _index(self, data):
        logger.debug(f' - ACTION: obj at index: `{data}`')
        try:
            if isinstance(data, dict):
                idx = int(data['value'])
                key = data['key']
            else:
                idx = int(data)
                key = 'default'
            result = storage.index(idx, key)
            self.send(f'index.{key}', {'index': idx, 'value': result})
        except IndexError as e:
            self._error_message(f'{e}')
        except ValueError as e:
            k = data
            if isinstance(data, dict):
                k = data['value']
            self._error_message(f'`{k}` is not a valid index.')

    def _remove(self, data):
        logger.debug(f' - ACTION: remove obj at index: `{data}`')
        try:
            if isinstance(data, dict):
                idx = int(data['value'])
                key = data['key']
            else:
                idx = int(data)
                key = 'default'
            result = storage.remove(idx, key)
            self.send(f'remove.{key}', {'index': idx, 'value': result})
            logger.debug(f'{storage.list(key)}')
        except KeyError as e:
            self._error_message(f'{e}')
        except IndexError as e:
            self._error_message(f'{e}')
        except ValueError as e:
            print(e.args)
            k = data
            if isinstance(data, dict):
                k = data['value']
            self._error_message(f'`{k}` is not a valid index.')

    def _error_message(self, msg):
        logger.warning(msg)
        self.error(msg)

    def _regist_handlers(self):
        self._client.on(PP_MESSAGE, self._message_handler)
        self._client.on('connect', self._client_connect)
        self._client.on('disconnect', self._client_disconnect)

    def _dynamic_message_handler(self, message: str, data):
        temp = message.split(self.separator)
        cmd = temp[0]
        key = temp[1]

        handler = self._handler_mapper.get(cmd, None)
        if handler:
            logger.info(f'Receive dynamic message from ProtoPie Connect.')
            logger.info(
                f'Message: `{cmd}`, command key: `{key}`. Data: `{data}`.')
            logger.debug(
                f'Handler: `{handler.__module__}.{handler.__name__}`.')

            return handler({
                'key': key,
                'value': data.get('value', None),
                'piedmont.payload': True
            })
        else:
            logger.debug(f'No handler for message: "{cmd}"')

    def _message_handler(self, data):
        msgId = data['messageId']
        logger.debug(f'Receive message: `{msgId}`')
        if len(msgId.split(self.separator)) > 1:
            logger.debug(f'Message `{msgId}` is a dynamic message.')
            return self._dynamic_message_handler(msgId, data)

        handler = self._handler_mapper.get(msgId, None)
        if handler:
            logger.info(f'Receive message from ProtoPie Connect.')
            logger.info(f'Message: `{msgId}`. Data: `{data}`.')
            logger.debug(f'Handler: `{handler.__name__}`.')
            return handler(data.get('value', None))
        else:
            logger.debug(f'No handler for message: "{msgId}"')

    def _client_disconnect(self, data: t.Any = None):
        print(
            f'Disconnect from: "{self._config.server}". {data or ""}')

    def _client_connect(self, data: t.Any = None):
        print(f'Connect to: "{self._config.server}". {data or ""}')
        self._client.emit(PP_BRIDGE_APP, {'name': self._config.app_name})

    def connect(self):
        try:
            self._client.connect(self._config.server)
        except socketio.exceptions.ConnectionError as e:
            logger.error(f'Opps. Error occurred when connecting to server.')
            logger.error(f'Error message: `{e}`.')
            logger.error(f'Please open `ProtoPie Connect` before start.')
            raise SystemExit(1)

    def bridge(
        self,
        messageId: str,
        is_dynamic: bool = False,
    ):
        def decorator(func):
            key = messageId
            if is_dynamic:
                key = messageId + self.separator

            self._regist_bridge_handler(key, func)

            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)
            return wrapper
        return decorator

    def _regist_bridge_handler(self, messageId: str, handler: T_Handler):
        old_func = self._handler_mapper.get(messageId, None)
        if old_func:
            raise DuplicateHandlerError(messageId)

        self._handler_mapper[messageId] = handler

    def send(self, messageId: str, value: t.Union[str, t.List[t.Any], t.Dict[t.AnyStr, t.Any]] = ""):

        if isinstance(value, str):
            data = value
        else:
            data = json.dumps(value)

        logger.info(f'Sending message to ProtoPie Connect.')
        logger.info(f'Message: `{messageId}`.')
        logger.info(f'Value: `{data}`')

        self._client.emit(
            PP_MESSAGE, {'messageId': messageId, 'value': data})

    def _clear(self, data):
        storage.clear('all')

    def __del__(self):
        if self._client.connected:
            self._client.disconnect()



from __future__ import annotations

import typing as t
from pprint import pprint
import json

from . import logger
from .errors import KeyPathError, StackNotExists


def _convert_key(key: str):
    if _is_valid_number_string(key):
        return int(key)

    return key


def _safe_load_json(data: any):
    try:
        data = json.loads(str(data))
    except json.decoder.JSONDecodeError:
        pass
    finally:
        return data


class BaseStorage:
    _stack: t.Dict[str, t.List[t.Any]] = {
        "default": []
    }
    _queue: t.Dict[str, t.List[t.Any]] = {
        "default": []
    }
    _list: t.Dict[str, t.List[t.Any]] = {
        'default': []
    }
    _dict: dict = {}

    def stack_count(self, key: str = 'default'):
        return len(self.stack(key))

    def queue_count(self, key: str = 'default'):
        return len(self.queue(key))

    def list_count(self, key: str = 'default'):
        return len(self.list(key))

    def dict_count(self):
        return len(self.dict.keys())

    def list(self, key: str = 'default') -> t.List[t.Any] | None:
        return self._list.get(key, None)

    def stack(self, key: str = 'default') -> t.List[t.Any] | None:
        return self._stack.get(key, None)

    def queue(self, key: str = 'default') -> t.List[t.Any] | None:
        return self._queue.get(key, None)

    @property
    def dict(self) -> t.Dict[str, t.Any]:
        return self._dict

    @property
    def keys(self) -> t.List[str]:
        return self.dict.keys()

    def append(self, obj: t.Any, key: str = 'default') -> t.List[t.Any]:
        data = _safe_load_json(obj)
        self._list[key] = self.list(key) or []
        self._list[key].append(data)
        return self.list(key)

    def insert(self, obj: any, index=0, key: str = 'default') -> t.List[t.Any]:
        data = _safe_load_json(obj)
        self._list[key] = self.list(key) or []
        self._list[key].insert(index, data)
        return self.list(key)

    def remove(self, index: int, key: str = 'default'):
        l = self.list(key)
        if l is None:
            raise KeyError(f'List `{key}` is not exists.')
        if abs(index) >= len(l):
            raise IndexError(
                f'Index `{index}` is out of range for list `{key}`')
        l.pop(index)

    def index(self, idx: int, key: str = 'default') -> any:
        self._list[key] = self.list(key) or []
        return self._list[key][idx]

    def push(
            self, obj: t.Any,
            target: t.Literal['stack', 'queue'] = 'stack',
            key: str = 'default'
    ) -> t.List[t.Any]:
        data = _safe_load_json(obj)
        if target == 'stack':
            self._stack[key] = self.stack(key) or []
            self._stack[key].append(data)
            return self.stack(key)
        else:
            self._queue[key] = self.queue(key) or []
            self._queue[key].append(data)
            return self.queue(key)

    def pop(
        self,
        target: t.Literal['stack', 'queue'] = 'stack',
        key: str = 'default',
        index: int = None
    ) -> any:
        if target == 'stack':
            index = index or -1
            logger.debug(f'Pop stack with name: `{key}`')
            if (stack := self.stack(key)) is not None:
                if index >= len(stack) and len(stack) > 0:
                    index = -1
                return stack.pop(index)
            raise StackNotExists(key)
        else:
            index = index or 0
            return self.queue(key).pop(index)

    def peek(
        self,
        target: t.Literal['stack', 'queue'] = 'stack',
        key: str = 'default'
    ) -> any:
        if target == 'stack':
            return self.stack(key)[-1]
        else:
            return self.queue(key)[0]

    def clear(self, target: t.Literal['stack', 'list', 'dict', 'queue', 'all'] = 'dict'):
        if target == 'stack':
            self._stack = {
                "default": []
            }
        elif target == 'list':
            self._list = {
                "default": []
            }
        elif target == 'queue':
            self._queue = {
                'default': []
            }
        elif target == 'dict':
            self._dict = {}
        else:
            self._stack = {
                "default": []
            }
            self._list = {
                "default": []
            }
            self._queue = {
                'default': []
            }
            self._dict = {}

    def show_dict(self):
        print('=' * 10, 'DICT DATA', '=' * 10)
        pprint(self.dict)

    def show_stack(self):
        print('=' * 10, 'STACK DATA', '=' * 10)
        pprint(self._stack)

    def show_list(self):
        print('=' * 10, 'LIST DATA', '=' * 10)
        pprint(self._list)

    def show_queue(self):
        print('=' * 10, 'QUEUE DATA', '=' * 10)
        pprint(self._queue)


class Storage(BaseStorage):

    def _set_value_by_key_path(
        self, key_path: list[str],
        value: any
    ):
        current = self.dict

        for subkey in key_path[:-1]:
            if _is_valid_number_string(subkey):
                current = current[int(subkey)]
            else:
                try:
                    current = current[subkey]
                    logger.debug(f'>>> {current}')
                except KeyError:
                    logger.debug(
                        f'Can not find key:`{subkey}`, create dict for this key.')
                    current[subkey] = {}
                    current = current[subkey]

        if isinstance(current, dict):
            current[key_path[-1]] = value
        elif isinstance(current, list):
            try:
                current[_convert_key(key_path[-1])] = value
            except IndexError:
                current.append(value)

        return value

    def set_value_by_key(
            self, key: str, value: any,
    ):
        key_path = key.split('.')

        if _is_valid_number_string(key_path[0]) or _is_valid_number_string(key):
            raise KeyPathError(key)
        data = _safe_load_json(value)

        if len(key_path) > 1:
            logger.debug(f'set value for key path: {key_path}')
            return self._set_value_by_key_path(key_path, data)
        else:
            self.dict[key] = data
            return data

    def get_value_by_key(self, key: str) -> str:
        key_path = key.split('.')
        try:
            if len(key_path) > 1:
                if _is_valid_number_string(key_path[0]):
                    raise KeyPathError(key)

                return self._get_value_by_key_path(key_path)
            else:
                return self.dict[key]
        except KeyError:
            raise KeyError(
                f'Can not find value for key: `{key}`.')

    def _get_value_by_key_path(self, key_path: list[str]) -> any:
        current = self.dict
        for key in key_path:
            if isinstance(current, list) and _is_valid_number_string(key):
                key = int(key)
            current = current[key]
        return current


def _is_valid_number_string(num: str) -> bool:
    try:
        temp = int(num)
        return True
    except ValueError:
        return False


storage = Storage()



from __future__ import annotations

import typing as t
from pathlib import Path

import yaml

from . import logger


class BaseConfig(dict):

    @property
    def app_name(self) -> str:
        return self['APP_NAME']

    @app_name.setter
    def app_name(self, value):
        self['APP_NAME'] = value

    @property
    def host(self):
        return self['HOST']

    @host.setter
    def host(self, value):
        self['HOST'] = str(value)

    @property
    def port(self):
        return self['PORT']

    @port.setter
    def port(self, value):
        self['PORT'] = str(value)

    @property
    def server(self):
        return f'{self.host}:{self.port}'


class FileConfig(BaseConfig):

    def __init__(self, path: Path):
        super().__init__()
        try:
            with open(path, 'r', encoding='utf-8') as f:
                temp = yaml.safe_load(f)
                logger.devlog(f'load config from file: {temp}')
                self.app_name = temp['name']
                self.host = temp['host']
                self.port = temp['port']
        except Exception as e:
            logger.warning(
                f'Exception occurred while parsing config file: {e}.\n'
                'Using default settings instead.'
            )
            self.update({
                "APP_NAME": 'Piedmont',
                'HOST': 'http://localhost',
                'PORT': '9981'
            })


class Config(BaseConfig):

    def __init__(
        self,
        name: str = None,
        host: str = None,
        port: str = None,
    ):
        super().__init__()
        self.app_name = name or 'Piedmont'
        self.host = host or 'http://localhost'
        self.port = port or '9981'



from piedmont.app import _key_index_parser


def test_insert_key_handelr():
    data = {
        'key': 'key1.key2'
    }
    key, idx = _key_index_parser(data)
    assert key == 'key1.key2'
    assert idx == 0

    data = {
        'key': 'key1.key2.1'
    }
    key, idx = _key_index_parser(data)
    assert key == 'key1.key2'
    assert idx == 1

    data = {
        'key': '3'
    }
    key, idx = _key_index_parser(data)
    assert key == 'default'
    assert idx == 3


from setuptools import setup, find_packages

setup(
    dependency_links=[],
    name="piedmont",
    version="0.2.0",
    description="A python library that helps you to build ProtoPie Bridge App or plugin more efficiently",
    long_description=open("README.md").read(),
    long_description_content_type=("text/markdown"),
    author="Nestor",
    author_email="admin@nestor.me",
    url="https://github.com/nestorrid/piedmont",
    packages=find_packages(),
    install_requires=[
        "bidict==0.23.1; python_version >= '3.8'",
        "certifi==2025.1.31; python_version >= '3.6'",
        "charset-normalizer==3.4.1; python_version >= '3.7'",
        "h11==0.14.0; python_version >= '3.7'",
        "idna==3.10; python_version >= '3.6'",
        "python-engineio==4.11.2; python_version >= '3.6'",
        "python-socketio==5.12.1; python_version >= '3.8'",
        "pyyaml==6.0.2; python_version >= '3.8'",
        "requests==2.32.3; python_version >= '3.8'",
        "simple-websocket==1.1.0; python_version >= '3.6'",
        "urllib3==2.3.0; python_version >= '3.9'",
        "websocket-client==1.8.0; python_version >= '3.8'",
        "wsproto==1.2.0; python_full_version >= '3.7.0'",
    ],
)
